@@ -1,5 +1,5 @@
 /*
-** Copyright 1998 - 2025 Double Precision, Inc.  See COPYING for
+** Copyright 1998 - 2025 S. Varshavchik.  See COPYING for
 ** distribution information.
 */
 
@@ -1335,6 +1335,10 @@ auto rfc2231_attr_encode(std::string_view name,
   the MIME object does not need rewriting, and true if autoconvert() will
   make changes.
 
+  rfc2045::entity::autoconvert_meta metadata;
+
+  metadata.appid="courier";
+
   rfc2045::entity::line_iter<false>::autoconvert(
       entity,
       []
@@ -1343,7 +1347,7 @@ auto rfc2231_attr_encode(std::string_view name,
          // ...
       },
       input_stream,
-      appid);
+      metadata);
 
   autoconvert() proceeds and rewrites the MIME entity, accordingly. The
   second parameter is a callable object that gets called repeatedly with
@@ -1355,9 +1359,16 @@ auto rfc2231_attr_encode(std::string_view name,
   that implements the same functions as std::streambuf. This streambuf
   represents the contents of the parsed mime entity.
 
-  The last parameter is optional. It's the name of the application that's
-  calling autoconvert(), this is recorded in the X-Mime-Autoconverted headers,
-  if any are added.
+  The last parameter is optional. It is an object with the following fields:
+
+  - appid: this is the name of the application that's calling autoconvert(),
+    this is recorded in the X-Mime-Autoconverted headers, if any are added.
+
+  - rwheader(): the default implementation does nothing, and returns its
+    header parameter value. A subclass can override and implement a function
+    that gets called for every logical header. The function may return a
+    different header, and this effectively replaces the original header in
+    the message.
 
   ENUMERATING MIME STRUCTURE
   ==========================
@@ -1427,6 +1438,7 @@ auto rfc2231_attr_encode(std::string_view name,
 #define RFC2045_ERRUNKNOWNTE		0x0100
 #define RFC2045_ERRINVALIDBASE64	0x0200
 #define RFC2045_ERRLONGQUOTEDPRINTABLE  0x0400
+#define RFC2045_ERRDUPLICATECONTENT	0x0800
 #define RFC2045_ERRFATAL		0x8000
 
 class rfc2045::entity_info {
@@ -1551,6 +1563,8 @@ class rfc2045::entity_info {
 
 	std::string_view content_type_charset() const;
 	std::string_view content_type_boundary() const;
+	std::string content_type_start() const;
+	const entity *content_type_multipart_signed() const;
 
 	// We don't need to look at these headers here,
 	// so we'll just grab them and parse them
@@ -1696,6 +1710,31 @@ class rfc2045::entity : public entity_info {
 
 	static constexpr std::string_view boundary_msg[]=RFC2045MIMEARRAY;
 
+	// Last parameter to autoconvert()
+
+	struct autoconvert_meta {
+		autoconvert_meta();
+		~autoconvert_meta();
+
+		std::string appid;
+
+		// The default implementation returns the full_header value
+		//
+		// The returned string_view's contents must exist at least
+		// until either the next call to rwheader(), or until
+		// autoconvert() returns.
+
+		virtual std::string_view rwheader(
+			// Mime entity whose header is getting passed in
+			const entity &e,
+
+			// Header's name in lowercase
+			std::string_view lcname,
+
+			// "Header: value<NEWLINE>"
+			std::string_view full_header);
+	};
+
 	// Factory for iterators and parsers that use LF(false) or CRLF(true)
 	// newline sequence.
 
@@ -1720,11 +1759,11 @@ class rfc2045::entity : public entity_info {
 			 typename src_type>
 		static auto autoconvert(const entity &e, out_iter &&closure,
 					src_type &src,
-					std::string_view appname="")
+					autoconvert_meta &metadata)
 		{
 			unsigned counter=0;
 
-			autoconvert(e, closure, src, counter, appname);
+			autoconvert(e, closure, src, counter, metadata);
 
 			if constexpr (std::is_same_v<out_iter, out_iter &>)
 			{
@@ -1736,14 +1775,15 @@ class rfc2045::entity : public entity_info {
 			 typename src_type>
 		static void autoconvert(const entity &e, out_iter &closure,
 					src_type &src, unsigned &counter,
-					std::string_view);
+					autoconvert_meta &metadata);
 
 		template<typename out_iter,
 			 typename src_type>
-		static void autoconvert_entity(const entity &e,
-					       out_iter &closure,
-					       src_type &src,
-					       std::string_view appname);
+		static void autoconvert_entity(
+			const entity &e,
+			out_iter &closure,
+			src_type &src,
+			autoconvert_meta &metadata);
 	};
 
 	entity() noexcept;
@@ -2493,6 +2533,15 @@ public:
 			line.insert(line.end(), bp, bp+l);
 		}
 
+		// Proactively disarm any practical jokes
+		line.erase(
+			std::remove_if(line.begin(), line.end(),
+				       [](char &c)
+				       {
+					       return c == 0 || c == '\r';
+				       }), line.end());
+
+
 		auto lb=line.begin(), le=line.end(),
 			p=std::find(lb, le, ':');
 
@@ -2504,8 +2553,7 @@ public:
 			{
 				return c != ':' &&
 					c != ' ' &&
-					c != '\t' &&
-					c != '\r';
+					c != '\t';
 			});
 
 		while (le > q)
@@ -2513,7 +2561,6 @@ public:
 			switch(le[-1]) {
 			case ' ':
 			case '\t':
-			case '\r':
 				--le;
 				continue;
 			}
@@ -2535,6 +2582,7 @@ void rfc2045::entity::parse(line_iter_type &iter)
 
 	std::string line;
 
+	bool duplicate_content=false;
 	while (iter.in_header())
 	{
 		auto [name, header]=iter.next_folded_header_line(*this, line);
@@ -2553,6 +2601,9 @@ void rfc2045::entity::parse(line_iter_type &iter)
 		{
 			content_type=rfc2231_header{header};
 
+			if (has_content_type_header)
+				duplicate_content=true;
+
 			has_content_type_header=true;
 
 			content_type.lowercase_value("charset");
@@ -2560,6 +2611,9 @@ void rfc2045::entity::parse(line_iter_type &iter)
 		}
 		if (name == "content-transfer-encoding")
 		{
+			if (has_content_transfer_encoding)
+				duplicate_content=true;
+
 			content_transfer_encoding=to_cte(
 				header.size() ? *header.data():0);
 
@@ -2625,8 +2679,18 @@ void rfc2045::entity::parse(line_iter_type &iter)
 				  RFC2045_ERRFATAL);
 		return;
 	}
-	else if (mime1)
+	else if (!mime1)
 	{
+		content_type=rfc2231_header{"text/plain"};
+		content_transfer_encoding=cte::sevenbit;
+	}
+	else
+	{
+		if (duplicate_content)
+		{
+			iter.report_error(RFC2045_ERRDUPLICATECONTENT |
+					  RFC2045_ERRFATAL);
+		}
 		if (rfc2045_message_content_type(content_type.value.c_str()) &&
 		    (content_transfer_encoding == cte::sevenbit ||
 		     content_transfer_encoding == cte::eightbit))
@@ -3224,7 +3288,7 @@ void rfc822::mime_decoder<out_iter, src_type>::decode(
 
 	if (decode_header)
 	{
-		typename rfc2045::entity::line_iter<crlf>::headers<
+		typename rfc2045::entity::line_iter<crlf>::template headers<
 			src_type> parser{e, src	};
 
 		parser.name_lc=header_name_lc;
@@ -3477,13 +3541,21 @@ std::string rfc2045::entity::new_boundary(src_type &&src, unsigned &counter)
 
 template<bool crlf>
 template<typename out_iter, typename src_type>
-void rfc2045::entity::line_iter<crlf>
-::autoconvert(const entity &e, out_iter &closure,
-	      src_type &src, unsigned &counter, std::string_view appname)
+void rfc2045::entity::line_iter<crlf>::autoconvert(
+	const entity &e, out_iter &closure,
+	src_type &src, unsigned &counter, autoconvert_meta &metadata)
 {
-	if (e.subentities.empty())
+	if (e.subentities.empty()
+
+	    // If some subentity needs to be rewritten then autoconvert_check
+	    // set rewrite_transfer_encoding from its default cte:error value
+	    // to cte::8bit, otherwise go into autoconvert_entity in order to
+	    // preserve the entire multipart entity, including the boundary
+	    // delimiter, and avoid needless work.
+
+	    || (e.multipart() && e.rewrite_transfer_encoding == cte::error))
 	{
-		autoconvert_entity(e, closure, src, appname);
+		autoconvert_entity(e, closure, src, metadata);
 		return;
 	}
 
@@ -3507,6 +3579,13 @@ void rfc2045::entity::line_iter<crlf>
 			{
 				auto current_header=
 					existing_headers.current_header();
+
+				current_header=metadata.rwheader(
+					e,
+					name_lc,
+					current_header
+				);
+
 				closure(current_header.data(),
 					current_header.size());
 			}
@@ -3517,7 +3596,7 @@ void rfc2045::entity::line_iter<crlf>
 		if (!e.subentities.empty()) // Sanity check
 		{
 			autoconvert(e.subentities[0], closure, src, counter,
-				    appname);
+				    metadata);
 		}
 		return;
 	}
@@ -3615,7 +3694,7 @@ void rfc2045::entity::line_iter<crlf>
 		separator += new_boundary;
 		separator += eol;
 		closure(separator.data(), separator.size());
-		autoconvert(subentity, closure, src, counter, appname);
+		autoconvert(subentity, closure, src, counter, metadata);
 		separator.clear();
 	}
 
