@@ -1,5 +1,5 @@
 /*
-** Copyright 2010-2014 Double Precision, Inc.  See COPYING for
+** Copyright 2010-2014 S. Varshavchik.  See COPYING for
 ** distribution information.
 */
 
@@ -53,7 +53,7 @@ static void doconvtoutf8_write_noeol(const char *p, size_t n,
 	}
 }
 
-static void doconvtoutf8_error(const char *p, int n,
+static void doconvtoutf8_error(const char *p, size_t n,
 			       void *void_arg)
 {
 	struct doconvtoutf8_info *ptr=(struct doconvtoutf8_info *)void_arg;
@@ -79,7 +79,7 @@ static int doconvtoutf8_rfc822hdr(const char *header,
 		doconvtoutf8_write(header, strlen(header), &info);
 		doconvtoutf8_write(": ", 2, &info);
 	}
-	rfc822_display_hdrvalue(header, value, "utf-8", 
+	rfc822_display_hdrvalue(header, value, "utf-8",
 				doconvtoutf8_write_noeol,
 				doconvtoutf8_error,
 				&info);
@@ -129,9 +129,17 @@ int rfc2045_decodemsgtoutf8(struct rfc2045src *src,
 			if (callback->flags & RFC2045_DECODEMSG_NOHEADERS)
 				continue;
 
+			if ((callback->flags &
+			     RFC2045_DECODEMSG_NOATTACHHEADERS)
+			    && p->parent)
+				continue;
+
 			if (doconvtoutf8_rfc822hdr(header, value,
 						   callback) < 0)
+			{
+				rfc2045header_end(hi);
 				return -1;
+			}
 		}
 		rfc2045header_end(hi);
 	}
